package ch.hepia.tpscala.monoid

import ch.hepia.tpscala.monoid

trait Monoid[A] {
  def op( a1: A, a2: A ): A
  def zero: A
}

object Monoid {
  def apply[A]( z: A )( f: (A,A)=>A ) = new Monoid[A] {
    def op( a1: A, a2: A ): A = f(a1,a2)
    def zero: A = z
  }

  object Laws {
    def checkAssociativity[A](
      mon: Monoid[A], a1: A, a2: A, a3: A
    ) = {
      import mon.op
      op(a1,op(a2,a3)) == op(op(a1,a2),a3)
    }
    def checkNeutral[A](
      mon: Monoid[A], a: A
    ) = {
      import mon._
      op(zero,a) == a && op(a,zero) == a
    }
  }
}

object Monoids {

  val intSum = Monoid(0)( _ + _ )
  val intProd = Monoid(1)( _ * _ )

  val doubleSum = Monoid( 0.0 )( _ + _ )
  val doubleMax = Monoid( Double.MinValue )( _ max _ )
  val doubleMin = Monoid( Double.MaxValue )( _ min _ )

  def listFlat[A] = Monoid[List[A]]( Nil )( _ ++ _ )

  def setUnion[A] = Monoid( Set.empty[A] )( _ ++ _ )

  def pipeline[A] = Monoid[A=>A]( identity )( _ andThen _ )

  def optionM[A]( op: (A,A)=>A ) =
    Monoid[Option[A]]( None ){ (lhs,rhs) =>
      (lhs,rhs) match {
        case (None,_) => rhs
        case (_,None) => lhs
        case (Some(a),Some(b)) => Some( op(a,b) )
      }
    }

  val optIntProd = optionM( intProd.op )
  val optDoubleMax = optionM( doubleMax.op )


  def tuple2M[A,B]( MA: Monoid[A], MB: Monoid[B] ) =
    Monoid[(A,B)]( (MA.zero,MB.zero) ){ (ab1,ab2) =>
      ( MA.op( ab1._1, ab2._1 ), MB.op( ab1._2, ab2._2 ) )
    }


  def mapM[K,V]( M: Monoid[V] ) = {
    def merge( lhs: Map[K,V], rhs: Map[K,V] ): Map[K,V] = {
      val ks = lhs.keySet ++ rhs.keySet
      ks.map{ k =>
        val v = M.op(
          lhs.getOrElse(k,M.zero),
          rhs.getOrElse(k,M.zero) )
        k -> v
      }.toMap
    }
    Monoid[Map[K,V]]( Map() )( merge )
  }
}

object Examples {

  import Monoids._

  def mapReduce[A,B]( as: List[A] )(f: A=>B, M: Monoid[B]): B = {
    def rec( rem: List[A], acc: B ): B = rem match {
      case Nil => acc
      case h :: t => rec( t, M.op(acc, f(h) ) )
    }
    rec( as, M.zero )
  }
  def reduce[A]( as: List[A] )( M: Monoid[A] ): A =
    mapReduce( as )( identity, M )

  //Donné à titre d'exemple
  def averageM[A]( MA: Monoid[A] ): Monoid[(A,Int)] =
    tuple2M( MA, intSum )
  val avgDouble = averageM( doubleSum )
  def average( as: List[Double] ): Option[Double] =
    if( as.isEmpty ) None
    else {
      val (sum,n) =
        mapReduce( as )( a => (a,1), avgDouble )
      Some( sum/n )
    }


  //Série 3: and
  def and( xs: List[Boolean] ): Boolean = reduce(xs)(Monoid[Boolean](true)(_ && _))

  //Retourne le min et le max d'une liste de doubles en un seul parcours
  def minMax( as: List[Double] ): (Double,Double) = mapReduce[Double, (Double, Double)](as)((a:Double) => (a, a), tuple2M(doubleMin, doubleMax))

  //Série 3: Even
  def even[A]( as: List[A] ): Boolean = mapReduce(as)(a=>false, Monoid(true)(_ == _))

  //Compte le nombre d'occurence de chaque objet de la liste
  def count[A]( as: List[A] ): Map[A,Int] = mapReduce[A, Map[A, Int]](as)((x: A) => Map[A, Int](x -> 1), mapM[A, Int](intSum))

  //Compte le nombre d'occurence de chaque pair d'objet
  def count2[A,B]( as: List[(A,B)] ): Map[A,Map[B,Int]] = mapReduce[(A,B), Map[A,Map[B, Int]]](as)( (x:(A,B)) => Map(x._1 -> Map(x._2 -> 1)),mapM[A,Map[B, Int]](mapM[B, Int](intSum)))

  //Série 4: forall
  def funcM[A,B](MB: Monoid[B]) = {
    Monoid[A=>B](a => MB.zero){ (f:A=>B, g:A=>B) => (a:A) => MB.op(f(a), g(a))}
  }

  def forall[A]( ps: List[(A)=>Boolean] ): (A)=>Boolean = reduce(ps)(funcM(Monoid[Boolean](true)(_ && _)))


  def sortedList[A](as1:List[A], as2:List[A], smallerThan: (A,A)=>Boolean) = {
    def mergeR(l1: List[A], l2: List[A], res:List[A]): List[A] = (l1,l2) match {
      case (Nil, _) => res.reverse ++ l2
      case (_, Nil) => res.reverse ++ l1
      case (h1::t1, h2::t2) =>
        if(smallerThan(h1,h2)) mergeR(t1, l2, h1::res)
        else mergeR(l1,t2, h2::res)
    }
    mergeR(as1, as2, Nil)
  }

  //def sort( as: List[Double]): List[Double] =
}



