import Monoid.Laws

trait Monoid[A] {
  def op( a1: A, a2: A ): A
  def zero: A
}

object Monoid {
  def apply[A]( z: A )( f: (A,A)=>A ) = new Monoid[A] {
    def op( a1: A, a2: A ): A = f(a1,a2)
    def zero: A = z
  }

  object Laws {
    def checkAssociativity[A](mon: Monoid[A], a1: A, a2: A, a3: A) = {
      import mon.op
      op(a1,op(a2,a3)) == op(op(a1,a2),a3)
    }
    def checkNeutral[A](mon: Monoid[A], a: A) = {
      import mon._
      op(zero,a) == a && op(a,zero) == a
    }
  }
}

case class Module[+A](name: String, args: Map[String, A]){
  override def toString: String = {
    s"Name : $name. Args : ${args.toString()}"
  }
}

case class Task(modules: List[Module[Any]]){
  override def toString: String = {
    s"Modules : ${modules.toString()}\n"
  }
}

object TaskMonoid{
  def + = Monoid[Task](Task(Nil))( (t1, t2) =>
    Task(t1.modules:::t2.modules))

  def mapReduce[A,B]( as: List[A] )(f: A=>B, M: Monoid[B]): B = {
    def rec( rem: List[A], acc: B ): B = rem match {
      case Nil => acc
      case h :: t => rec( t, M.op(acc, f(h) ) )
    }
    rec( as, M.zero )
  }
  def reduce[A]( as: List[A] )( M: Monoid[A] ): A =
    mapReduce( as )( identity, M )
}

object tp {
  def main(args: Array[String]): Unit = {
    import TaskMonoid._

    val m1 = Module("Swiss-Impex", Map("year" -> 2019, "month" -> 1, "category" -> "2709.0090"))
    val m2 = Module("Fao Genève PROD 2019/11/06", Map("passCaptcha" -> true, "page" -> 1))
    val m3 = Module("Tâche 3", Map("Oui" -> true, "non" -> false))

    val t1 = Task((m1::Nil))
    val t2 = Task((m2::Nil))
    val t3 = Task((m3::Nil))

    val t1p2 = reduce(List(t1, t2))(+)
    val t1t2t3 = List(t1, t2, t3)
    val t1t2 = List(t1, t2)
    val t1t3 = List(t1, t3)
    val t2t3 = List(t2,t3)


    println("========== TEST ASSOCIATIVITE ===========\n")
    println((reduce(List(t1, reduce(t2t3)(+)))(+) == reduce(List(reduce(t1t2)(+), t3))(+)))
    println()
    println("========== TEST ELEMENT NEUTRE ==========\n")
    println(t1 == reduce(List(t1,Task(Nil)))(+) && t1 == reduce(List(Task(Nil), t1))(+))

  }
}